import numpy as np


class ImageArray:
    def __init__(self, ndarray: np.ndarray) -> None:
        self.original_array = ndarray
        self.R = self.original_array[..., 0]
        self.G = self.original_array[..., 1]
        self.B = self.original_array[..., 2]
        self.A = None

        if self.original_array.shape[-1] == 4:
            self.A = self.original_array[..., 3]

    @property
    def has_alpha(self):
        return True if self.A is not None else False

    def constrain_channels(self) -> None:
        new_array = np.array([self.R, self.G, self.B], dtype=np.int16)
        np.clip(new_array, 0, 255, out=new_array)
        self.R, self.G, self.B = new_array

    def get_current(self) -> np.ndarray:
        """Return an array with the original shape and updated RGB(A) values."""
        new_array = np.empty(self.original_array.shape, dtype=np.uint8)
        new_array[..., 0] = self.R
        new_array[..., 1] = self.G
        new_array[..., 2] = self.B
        if self.has_alpha:
            new_array[..., 3] = self.A
        return new_array

    def get_current_rgb(self) -> np.ndarray:
        """Return an array with only RGB values."""
        rgb_shape = self.original_array.shape[:-1] + (3,)
        new_array = np.empty(rgb_shape, dtype=np.uint8)
        new_array[..., 0] = self.R
        new_array[..., 1] = self.G
        new_array[..., 2] = self.B
        return new_array



"""Presets."""


from fimage.filters import (
    Brightness,
    Clip,
    Colorize,
    Contrast,
    Curves,
    Exposure,
    Gamma,
    Grayscale,
    Posterize,
    Saturation,
    Vibrance,
)
from fimage.image_array import ImageArray


class Preset:
    @classmethod
    def process(cls, image_array: ImageArray) -> None:
        for transformation in cls.transformations:
            transformation.process(image_array)


class Love(Preset):
    transformations = [
        Brightness(5),
        Exposure(8),
        Contrast(4),
        Colorize(196, 32, 7, 30),
        Vibrance(50),
        Gamma(1.3),
    ]


class OrangePeel(Preset):
    transformations = [
        Curves((0, 0), (100, 50), (140, 200), (255, 255)),
        Vibrance(-30),
        Saturation(-30),
        Colorize(255, 144, 0, 30),
        Contrast(-5),
        Gamma(1.4),
    ]


class SinCity(Preset):
    transformations = [
        Contrast(100),
        Vibrance(15),
        Exposure(10),
        Posterize(80),
        Clip(30),
        Grayscale(),
    ]



from .converters import hsv2rgb, rgb2hsv
from .filters import *
from .fimage import FImage
from .image_array import ImageArray
from .presets import *

__all__ = [
    "rgb2hsv",
    "hsv2rgb",
    "FImage",
    "ImageArray",
    "Brightness",
    "Channels",
    "Clip",
    "Colorize",
    "Contrast",
    "Curves",
    "Exposure",
    "Gamma",
    "Grayscale",
    "FillColor",
    "Hue",
    "Invert",
    "Noise",
    "Posterize",
    "Saturation",
    "Sepia",
    "Vibrance",
    "Love",
    "OrangePeel",
    "SinCity",
]



import numpy as np
from PIL import Image, ImageOps

from fimage.image_array import ImageArray


class FImage:
    def __init__(self, image) -> None:
        self.original_image = Image.open(image)
        # Correct image orientation based on exif information
        self.image = ImageOps.exif_transpose(self.original_image)
        self.exif_data = self.image.getexif()
        self.image_array = ImageArray(np.array(self.image))

    def apply(self, *filters):
        for t in filters:
            t.process(self.image_array)

        self.image = self.ndarray_to_image()

    def ndarray_to_image(self):
        return Image.fromarray(self.image_array.get_current(), self.image.mode)

    def save(self, *args, **kwargs):
        self.image.save(exif=self.exif_data, *args, **kwargs)



"""Color filters."""


import abc
import math
from typing import Dict, Tuple

import numpy as np

from fimage.converters import hsv2rgb, rgb2hsv
from fimage.image_array import ImageArray


class Filter(abc.ABC):
    @abc.abstractmethod
    def process(self, image_array: ImageArray) -> None:
        pass


class FillColor(Filter):
    def __init__(self, R: int, G: int, B: int) -> None:
        self.R = R
        self.G = G
        self.B = B

    def process(self, image_array: ImageArray) -> None:
        image_array.R = self.R
        image_array.G = self.G
        image_array.B = self.B


class Sepia(Filter):
    def __init__(self, adjust: int = 100) -> None:
        self.adjust = adjust / 100

    def process(self, image_array: ImageArray) -> None:
        image_array.R = (
            image_array.R * (1 - (0.607 * self.adjust))
            + (image_array.G * (0.769 * self.adjust))
            + (image_array.B * (0.189 * self.adjust))
        )
        image_array.G = (
            (image_array.R * (0.349 * self.adjust))
            + (image_array.G * (1 - (0.314 * self.adjust)))
            + (image_array.B * (0.168 * self.adjust))
        )
        image_array.B = (
            (image_array.R * (0.272 * self.adjust))
            + (image_array.G * (0.534 * self.adjust))
            + (image_array.B * (1 - (0.869 * self.adjust)))
        )
        image_array.constrain_channels()


class Contrast(Filter):
    def __init__(self, adjust: int = 100) -> None:
        self.adjust = math.pow((adjust + 100) / 100, 2)

    def process(self, image_array: ImageArray) -> None:
        image_array.R = image_array.R / 255
        image_array.R -= 0.5
        image_array.R *= self.adjust
        image_array.R += 0.5
        image_array.R *= 255

        image_array.G = image_array.G / 255
        image_array.G -= 0.5
        image_array.G *= self.adjust
        image_array.G += 0.5
        image_array.G *= 255

        image_array.B = image_array.B / 255
        image_array.B -= 0.5
        image_array.B *= self.adjust
        image_array.B += 0.5
        image_array.B *= 255
        image_array.constrain_channels()


class Brightness(Filter):
    def __init__(self, adjust: int = 0) -> None:
        self.adjust = math.floor(255 * (adjust / 100))

    def process(self, image_array: ImageArray) -> None:
        ndarray = image_array.get_current().astype(np.int16)
        ndarray += self.adjust
        image_array.R = ndarray[..., 0]
        image_array.G = ndarray[..., 1]
        image_array.B = ndarray[..., 2]
        image_array.constrain_channels()


class Saturation(Filter):
    def __init__(self, adjust: int = 0) -> None:
        self.adjust = adjust * -0.01

    def process(self, image_array: ImageArray) -> None:
        ndarray = image_array.get_current()
        axis = ndarray.ndim - 1

        max_array = ndarray.max(axis=axis)
        max_array = np.repeat(max_array, 3).reshape(ndarray.shape)

        sat_array = np.where(
            ndarray != max_array,
            ndarray + (max_array - ndarray) * self.adjust,
            ndarray,
        )
        image_array.R = sat_array[..., 0]
        image_array.G = sat_array[..., 1]
        image_array.B = sat_array[..., 2]
        image_array.constrain_channels()


class Vibrance(Filter):
    def __init__(self, adjust: int = 0) -> None:
        self.adjust = adjust * -1

    def process(self, image_array: ImageArray) -> None:
        ndarray = image_array.get_current_rgb()

        axis = ndarray.ndim - 1

        max_array = ndarray.max(axis=axis)
        max_array = np.repeat(max_array, 3).reshape(ndarray.shape)

        avg_array = np.mean(ndarray, axis=axis, dtype=np.uint8)
        avg_array = np.repeat(avg_array, 3).reshape(ndarray.shape)

        amt_array = (((max_array - avg_array) * 2 / 255) * self.adjust) / 100

        vib_array = np.where(
            ndarray != max_array,
            ndarray + (max_array - ndarray) * amt_array,
            ndarray,
        )
        image_array.R = vib_array[..., 0]
        image_array.G = vib_array[..., 1]
        image_array.B = vib_array[..., 2]
        image_array.constrain_channels()


class Grayscale(Filter):
    def process(self, image_array: ImageArray) -> None:
        avg = (
            (0.299 * image_array.R) + (0.587 * image_array.G) + (0.114 * image_array.B)
        )
        image_array.R = avg
        image_array.G = avg
        image_array.B = avg


class Hue(Filter):
    def __init__(self, adjust: int = 0) -> None:
        self.adjust = abs(adjust)

    def process(self, image_array: ImageArray) -> None:
        ndarray = image_array.get_current_rgb()
        hsv = rgb2hsv(ndarray)
        h = hsv[..., 0]
        h *= 100
        h += self.adjust
        h = h % 100
        h = h / 100
        hsv[..., 0] = h
        # hsv2rgb function returns a float type with values ranging from 0 to 1
        rgb = (hsv2rgb(hsv) * 255).astype(np.uint8)
        image_array.R = rgb[..., 0]
        image_array.G = rgb[..., 1]
        image_array.B = rgb[..., 2]
        image_array.constrain_channels()


class Colorize(Filter):
    def __init__(self, R: int, G: int, B: int, level: int) -> None:
        self.R = R
        self.G = G
        self.B = B
        self.level = level

    def process(self, image_array: ImageArray) -> None:
        image_array.R = image_array.R - (image_array.R - self.R) * (self.level / 100)
        image_array.G = image_array.G - (image_array.G - self.G) * (self.level / 100)
        image_array.B = image_array.B - (image_array.B - self.B) * (self.level / 100)
        image_array.constrain_channels()


class Invert(Filter):
    def process(self, image_array: ImageArray) -> None:
        image_array.R = 255 - image_array.R
        image_array.G = 255 - image_array.G
        image_array.B = 255 - image_array.B
        image_array.constrain_channels()


class Gamma(Filter):
    def __init__(self, adjust: int = 1) -> None:
        self.adjust = adjust

    def process(self, image_array: ImageArray) -> None:
        image_array.R = ((image_array.R / 255) ** self.adjust) * 255
        image_array.G = ((image_array.G / 255) ** self.adjust) * 255
        image_array.B = ((image_array.B / 255) ** self.adjust) * 255
        image_array.constrain_channels()


class Noise(Filter):
    def __init__(self, adjust: int = 1) -> None:
        self.adjust = abs(adjust) * 2.55

    def process(self, image_array: ImageArray) -> None:
        # Generate a random array and sum it up with our current one
        random_array = np.random.randint(
            -self.adjust, self.adjust, image_array.original_array.shape
        )
        ndarray = image_array.get_current()
        ndarray = ndarray + random_array

        image_array.R = ndarray[..., 0]
        image_array.G = ndarray[..., 1]
        image_array.B = ndarray[..., 2]

        image_array.constrain_channels()


class Clip(Filter):
    def __init__(self, adjust: int = 1) -> None:
        self.adjust = abs(adjust) * 2.55

    def process(self, image_array: ImageArray) -> None:
        ndarray = image_array.get_current()

        ndarray = np.where(ndarray > 255 - self.adjust, 255, ndarray)
        ndarray = np.where(ndarray < self.adjust, 0, ndarray)

        image_array.R = ndarray[..., 0]
        image_array.G = ndarray[..., 1]
        image_array.B = ndarray[..., 2]


class Channels(Filter):
    def __init__(self, channels: Dict = {}) -> None:
        self.R = channels.get("R")
        self.G = channels.get("G")
        self.B = channels.get("B")

        if self.R:
            self.R /= 100

        if self.G:
            self.G /= 100

        if self.B:
            self.B /= 100

    def process(self, image_array: ImageArray) -> None:
        if self.R is not None:
            if self.R > 0:
                image_array.R = image_array.R + (255 - image_array.R) * self.R
            else:
                image_array.R = image_array.R + (255 - image_array.R) * abs(self.R)

        if self.G is not None:
            if self.G > 0:
                image_array.G = image_array.G + (255 - image_array.G) * self.G
            else:
                image_array.G = image_array.G + (255 - image_array.G) * abs(self.G)

        if self.B is not None:
            if self.B > 0:
                image_array.B = image_array.B + (255 - image_array.B) * self.B
            else:
                image_array.B = image_array.B + (255 - image_array.B) * abs(self.B)

        image_array.constrain_channels()


class Curves(Filter):
    def __init__(self, p0: Tuple, p1: Tuple, p2: Tuple, p3: Tuple) -> None:
        self.p0 = p0
        self.p1 = p1
        self.p2 = p2
        self.p3 = p3
        self.curve = self.calculate_bezier(1000)

    def clip(self, x: int) -> int:
        if x < 0:
            return 0

        if x > 255:
            return 255

        return x

    def missing_values(self, curve: Dict) -> Dict:
        max_value = 256

        if len(curve.keys()) < max_value:
            # Some values might be missed, we need to add them
            right_value = None
            left_value = None

            for index in range(max_value):
                if index not in curve.keys():
                    left_value = curve.get(index - 1)
                    for i in range(index, max_value):
                        if i in curve.keys():
                            right_value = curve.get(i)
                            break

                    if left_value is not None and right_value is not None:
                        curve[index] = round((left_value + right_value) / 2)

                    if left_value is None:
                        curve[index] = right_value

                    if right_value is None:
                        curve[index] = left_value

        return curve

    def calculate_bezier(self, granuality: int) -> Dict:
        result = dict()

        for i in range(granuality):
            t = i / granuality
            x = (
                (1 - t) ** 3 * self.p0[0]
                + 3 * (1 - t) ** 2 * t * self.p1[0]
                + 3 * (1 - t) * t ** 2 * self.p2[0]
                + t ** 3 * self.p3[0]
            )
            y = (
                (1 - t) ** 3 * self.p0[1]
                + 3 * (1 - t) ** 2 * t * self.p1[1]
                + 3 * (1 - t) * t ** 2 * self.p2[1]
                + t ** 3 * self.p3[1]
            )

            result[round(x)] = round(self.clip(y))

        result = self.missing_values(result)

        return result

    def process(self, image_array: ImageArray) -> None:
        # https://stackoverflow.com/a/16993364
        ndarray = image_array.get_current()

        u, inv = np.unique(ndarray, return_inverse=True)
        ndarray = np.array([self.curve.get(x) for x in u])[inv].reshape(ndarray.shape)

        image_array.R = ndarray[..., 0]
        image_array.G = ndarray[..., 1]
        image_array.B = ndarray[..., 2]


class Exposure(Filter):
    def __init__(self, adjust: int = 1) -> None:
        self.adjust = abs(adjust) / 100
        self.p1 = (0, self.adjust * 255)
        self.p2 = (255 - (self.adjust * 255), 255)

    def process(self, image_array: ImageArray) -> None:
        Curves((0, 0), self.p1, self.p2, (255, 255)).process(image_array)


class Posterize(Filter):
    def __init__(self, adjust) -> None:
        self.num_areas = 256 / adjust
        self.num_values = 255 / (adjust - 1)

    def process(self, image_array: ImageArray) -> None:
        ndarray = image_array.get_current()
        ndarray = np.floor(ndarray / self.num_areas) * self.num_values

        image_array.R = ndarray[..., 0]
        image_array.G = ndarray[..., 1]
        image_array.B = ndarray[..., 2]
        image_array.constrain_channels()



import numpy as np


def rgb2hsv(rgb: np.ndarray) -> np.ndarray:
    axis = rgb.ndim - 1

    rgb = rgb.astype(np.float64) / 255
    r = rgb[..., 0]
    g = rgb[..., 1]
    b = rgb[..., 2]

    max_array = rgb.max(axis=axis)
    min_array = rgb.min(axis=axis)

    v = max_array
    d = max_array - min_array

    old_settings = np.seterr(invalid="ignore")
    d[d == 0] = 0

    s = np.where(max_array == 0, 0, d / max_array)

    h = np.zeros_like(s)

    np.putmask(
        h,
        max_array == r,
        ((g - b) / d) + np.where(g < b, 6, 0).astype(np.float64),
    )
    np.putmask(h, max_array == g, (b - r) / d + 2)
    np.putmask(h, max_array == b, (r - g) / d + 4)

    np.seterr(**old_settings)

    h = h / 6
    hsv = np.empty_like(rgb)
    hsv[..., 0] = h
    hsv[..., 1] = s
    hsv[..., 2] = v
    hsv[np.isnan(hsv)] = 0
    return hsv


def hsv2rgb(hsv: np.ndarray) -> np.ndarray:
    input_shape = hsv.shape
    hsv = hsv.reshape(-1, 3)
    h, s, v = hsv[:, 0], hsv[:, 1], hsv[:, 2]

    i = np.int32(h * 6.0)
    f = (h * 6.0) - i
    p = v * (1.0 - s)
    q = v * (1.0 - s * f)
    t = v * (1.0 - s * (1.0 - f))
    i = i % 6

    rgb = np.zeros_like(hsv)
    v = v.reshape(-1, 1)
    t = t.reshape(-1, 1)
    p = p.reshape(-1, 1)
    q = q.reshape(-1, 1)

    rgb[i == 0] = np.hstack([v, t, p])[i == 0]
    rgb[i == 1] = np.hstack([q, v, p])[i == 1]
    rgb[i == 2] = np.hstack([p, v, t])[i == 2]
    rgb[i == 3] = np.hstack([p, q, v])[i == 3]
    rgb[i == 4] = np.hstack([t, p, v])[i == 4]
    rgb[i == 5] = np.hstack([v, p, q])[i == 5]
    rgb[s == 0.0] = np.hstack([v, v, v])[s == 0.0]

    return rgb.reshape(input_shape)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fimage']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.0.0,<9.0.0', 'numpy>=1.21.4,<2.0.0']

setup_kwargs = {
    'name': 'fimage',
    'version': '0.1.8',
    'description': 'A Python module to create and apply filters to images.',
    'long_description': '# FImage\n\nA Python module to apply and create multiples filters to images.\n\nYou need to be using Python 3.7 or greater to be able to use **FImage**.\n\n## Installation\n```python\npip install fimage\n```\n\n## Example\n\n### A Simple filter\n\nCreate a file `app.py`  with:\n\n```python\nfrom fimage import FImage\nfrom fimage.filters import Sepia\n\n\ndef main():\n    # replace \'my_picture.jpg\' with the path to your image\n    image = FImage(\'my_picture.jpg\')\n\n    # apply the Sepia filter to the image\n    image.apply(Sepia(90))\n\n    # save the image with the applied filter\n    image.save(\'my_picture_sepia.jpg\')\n\n\nif __name__ == "__main__":\n    main()\n```\n\nNow, just run it :\n\n```python\npython app.py\n```\n\nThis is `my_picture.jpg` before the filter was applied\n\n<img alt="my_picture.jpg" src="examples/img/my_picture.jpg" width="400" height="500">\n\nAnd this is how new image `my_picture_sepia.jpg` looks like after the filter was applied\n\n<img alt="my_picture_sepia.jpg" src="examples/img/my_picture_sepia.jpg" width="400" height="500">\n\n**Note**:  *90 represents the adjustment value we want to use for applying a sepia tone to this picture, lower values will result an image with less sepia tone while higher values will give us an image with a notorious sepia tone.*\n\nMost of the filters **FImage** offers will need an adjustment value to be passed.\n\n### Applying multiple filters\n\n**FImage** offers more filters besides the Sepia one, even you can combine multiples filters to give a better look to your picture.\n\nModify the file `app.py` to import more filters from **FImage**\n\n```python\nfrom fimage import FImage\nfrom fimage.filters import Contrast, Brightness, Saturation\n\n\ndef main():\n    image = FImage(\'my_picture.jpg\')\n\n    # apply the mutiple filters to the image\n    image.apply(\n        Saturation(20),\n        Contrast(25),\n        Brightness(15)\n    )\n\n    # save the image with the applied filter\n    image.save(\'my_picture_mixed.jpg\')\n\n\nif __name__ == "__main__":\n    main()\n```\n\nWe run it by\n\n```python\npython app.py\n```\n\nAnd our new `my_picture_mixed.jpg` looks like\n\n<img alt="my_picture_mixed.jpg" src="examples/img/my_picture_mixed.jpg" width="400" height="500">\n\nThe order in which the filters are passed to the `apply` function matters, this is because the filters are applied in a sequential manner, so the next filter will be applied over the resultant image from the previous one.\n\nCurrently **FImage** supports the following filters:\n- **FillColor**\n- **Sepia**\n- **Contrast**\n- **Brightness**\n- **Saturation**\n- **Vibrance**\n- **Grayscale**\n- **Hue**\n- **Colorize**\n- **Invert**\n- **Gamma**\n- **Noise**\n- **Clip**\n- **Exposure**\n\n### Presets\n\nPresets are just the combinations of multiple filters with already defined adjustment values.\n\nLet’s change our `app.py` one more time to use the Presets\n```python\nfrom fimage import FImage\nfrom fimage.presets import SinCity\n\n\ndef main():\n    # replace \'my_picture.jpg\' with the path to your image\n    image = FImage(\'my_picture.jpg\')\n\n    # apply the SinCity preset to the image\n    image.apply(SinCity())\n\n    # save the image with the applied preset\n    image.save(\'my_picture_sincity.jpg\')\n\n\nif __name__ == "__main__":\n    main()\n```\n\n After we run it, we get our new  `my_picture_sincity.jpg`\n\n<img alt="my_picture_sincity.jpg" src="examples/img/my_picture_sincity.jpg" width="400" height="500">\n\nCurrently supported Presets:\n- **SinCity**\n- **OrangePeel**\n- **Love**\n\n### Custom Presets\nIf you like the look your picture got after testing different filters and want to store this combination for applying it to more pictures, you can create your own Preset by just extending the `Preset` Class and specifying these filters and their adjust values in it.\n\nIn our `app.py` let’s do\n\n```python\nfrom fimage import FImage\nfrom fimage.presets import Preset\nfrom fimage.filters import Contrast, Brightness, Saturation\n\n\n# Create my custom preset and specify the filters to apply\nclass MyOwnPreset(Preset):\n    transformations = [\n        Contrast(30),\n        Saturation(50),\n        Brightness(10),\n    ]\n\n\ndef main():\n    # replace \'my_picture.jpg\' with the path to your image\n    image = FImage(\'my_picture.jpg\')\n\n    # apply MyOwnPreset to the image\n    image.apply(MyOwnPreset())\n\n    # save the image with the applied preset\n    image.save(\'my_picture_custom.jpg\')\n\n\nif __name__ == "__main__":\n    main()\n```\n\nThe new `my_picture_custom.jpg`\n\n<img alt="my_picture_custom.jpg" src="examples/img/my_picture_custom.jpg" width="400" height="500">\n\nNow, in this way `MyOwnPreset` has the combination of filters you like and you can use to modify more pictures.',
    'author': 'Jordan Jimenez',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jordandjp/fimage',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
